__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import base64
import logging
import sys

PY3 = sys.version_info[0] == 3  # should now always be true, kept for plugins

import flask as _flask
import flask_login

import octoprint.server
import octoprint.timelapse
import octoprint.vendor.flask_principal as flask_principal
from octoprint.plugin import plugin_manager
from octoprint.settings import settings
from octoprint.util import deprecated, to_unicode

from . import flask, sockjs, tornado, watchdog  # noqa: F401


@deprecated(
    "API keys are no longer needed for anonymous access and thus this is now obsolete"
)
def enforceApiKeyRequestHandler():
    pass


apiKeyRequestHandler = deprecated(
    "apiKeyRequestHandler has been renamed to enforceApiKeyRequestHandler"
)(enforceApiKeyRequestHandler)


def loginFromApiKeyRequestHandler():
    """
    ``before_request`` handler for blueprints which creates a login session for the provided api key (if available)

    App session keys are handled as anonymous keys here and ignored.
    """
    try:
        if loginUserFromApiKey():
            _flask.g.login_via_apikey = True
    except InvalidApiKeyException:
        _flask.abort(403)


def loginFromAuthorizationHeaderRequestHandler():
    """
    ``before_request`` handler for creating login sessions based on the Authorization header.
    """
    try:
        if loginUserFromAuthorizationHeader():
            _flask.g.login_via_header = True
    except InvalidApiKeyException:
        _flask.abort(403)


class InvalidApiKeyException(Exception):
    pass


def loginUserFromApiKey():
    apikey = get_api_key(_flask.request)
    if not apikey:
        return False

    user = get_user_for_apikey(apikey)
    if user is None:
        # invalid API key = no API key
        return False

    return loginUser(user, login_mechanism="apikey")


def loginUserFromAuthorizationHeader():
    authorization_header = get_authorization_header(_flask.request)
    user = get_user_for_authorization_header(authorization_header)
    return loginUser(user, login_mechanism="authheader")


def loginUser(user, remember=False, login_mechanism=None):
    """
    Logs the provided ``user`` into Flask Login and Principal if not None and active

    Args:
            user: the User to login. May be None in which case the login will fail
            remember: Whether to set the ``remember`` flag on the Flask Login operation

    Returns: (bool) True if the login succeeded, False otherwise

    """
    if (
        user is not None
        and user.is_active
        and flask_login.login_user(user, remember=remember)
    ):
        flask_principal.identity_changed.send(
            _flask.current_app._get_current_object(),
            identity=flask_principal.Identity(user.get_id()),
        )
        if login_mechanism:
            _flask.session["login_mechanism"] = login_mechanism
        return True
    return False


def requireLoginRequestHandler():
    if _flask.request.endpoint.endswith(".static"):
        return

    if not octoprint.server.userManager.has_been_customized():
        return

    user = flask_login.current_user
    if user is None or user.is_anonymous or not user.is_active:
        _flask.abort(403)


def corsRequestHandler():
    """
    ``before_request`` handler for blueprints which sets CORS headers for OPTIONS requests if enabled
    """
    if _flask.request.method == "OPTIONS" and settings().getBoolean(
        ["api", "allowCrossOrigin"]
    ):
        # reply to OPTIONS request for CORS headers
        return optionsAllowOrigin(_flask.request)


def corsResponseHandler(resp):
    """
    ``after_request`` handler for blueprints for which CORS is supported.

    Sets ``Access-Control-Allow-Origin`` headers for ``Origin`` request header on response.
    """

    # Allow crossdomain
    allowCrossOrigin = settings().getBoolean(["api", "allowCrossOrigin"])
    if (
        _flask.request.method != "OPTIONS"
        and "Origin" in _flask.request.headers
        and allowCrossOrigin
    ):
        resp.headers["Access-Control-Allow-Origin"] = _flask.request.headers["Origin"]

    return resp


def csrfRequestHandler():
    """
    ``before_request`` handler for blueprints which checks for CRFS double token on
    relevant requests & methods.
    """
    from octoprint.server.util.csrf import validate_csrf_request

    if settings().getBoolean(["devel", "enableCsrfProtection"]):
        validate_csrf_request(_flask.request)


def csrfResponseHandler(resp):
    """
    ``after_request`` handler for updating the CSRF cookie on each response.
    """
    from octoprint.server.util.csrf import add_csrf_cookie

    return add_csrf_cookie(resp)


def noCachingResponseHandler(resp):
    """
    ``after_request`` handler for blueprints which shall set no caching headers
    on their responses.

    Sets ``Cache-Control``, ``Pragma`` and ``Expires`` headers accordingly
    to prevent all client side caching from taking place.
    """

    return flask.add_non_caching_response_headers(resp)


def noCachingExceptGetResponseHandler(resp):
    """
    ``after_request`` handler for blueprints which shall set no caching headers
    on their responses to any requests that are not sent with method ``GET``.

    See :func:`noCachingResponseHandler`.
    """

    if _flask.request.method == "GET":
        return flask.add_no_max_age_response_headers(resp)
    else:
        return flask.add_non_caching_response_headers(resp)


def optionsAllowOrigin(request):
    """
    Shortcut for request handling for CORS OPTIONS requests to set CORS headers.
    """

    resp = _flask.current_app.make_default_options_response()

    # Allow the origin which made the XHR
    resp.headers["Access-Control-Allow-Origin"] = request.headers["Origin"]
    # Allow the actual method
    resp.headers["Access-Control-Allow-Methods"] = request.headers[
        "Access-Control-Request-Method"
    ]
    # Allow for 10 seconds
    resp.headers["Access-Control-Max-Age"] = "10"

    # 'preflight' request contains the non-standard headers the real request will have (like X-Api-Key)
    customRequestHeaders = request.headers.get("Access-Control-Request-Headers", None)
    if customRequestHeaders is not None:
        # If present => allow them all
        resp.headers["Access-Control-Allow-Headers"] = customRequestHeaders

    return resp


def get_user_for_apikey(apikey):
    if apikey is not None:
        if apikey == settings().get(["api", "key"]):
            # master key was used
            return octoprint.server.userManager.api_user_factory()

        user = octoprint.server.userManager.find_user(apikey=apikey)
        if user is not None:
            # user key was used
            return user

        apikey_hooks = plugin_manager().get_hooks("octoprint.accesscontrol.keyvalidator")
        for name, hook in apikey_hooks.items():
            try:
                user = hook(apikey)
                if user is not None:
                    return user
            except Exception:
                logging.getLogger(__name__).exception(
                    "Error running api key validator "
                    "for plugin {} and key {}".format(name, apikey),
                    extra={"plugin": name},
                )
    return None


def get_user_for_remote_user_header(request):
    if not settings().getBoolean(["accessControl", "trustRemoteUser"]):
        return None

    header = request.headers.get(settings().get(["accessControl", "remoteUserHeader"]))
    if header is None:
        return None

    user = octoprint.server.userManager.findUser(userid=header)

    if user is None and settings().getBoolean(["accessControl", "addRemoteUsers"]):
        octoprint.server.userManager.addUser(
            header, settings().generateApiKey(), active=True
        )
        user = octoprint.server.userManager.findUser(userid=header)

    return user


def get_user_for_authorization_header(header):
    if not settings().getBoolean(["accessControl", "trustBasicAuthentication"]):
        return None

    if header is None:
        return None

    if not header.startswith("Basic "):
        # we currently only support Basic Authentication
        return None

    header = header.replace("Basic ", "", 1)
    try:
        header = to_unicode(base64.b64decode(header))
    except TypeError:
        return None

    name, password = header.split(":", 1)
    if not octoprint.server.userManager.enabled:
        return None

    user = octoprint.server.userManager.find_user(userid=name)
    if settings().getBoolean(
        ["accessControl", "checkBasicAuthenticationPassword"]
    ) and not octoprint.server.userManager.check_password(name, password):
        # password check enabled and password don't match
        return None

    return user


def get_api_key(request):
    # Check Flask GET/POST arguments
    if hasattr(request, "values") and "apikey" in request.values:
        return request.values["apikey"]

    # Check Tornado GET/POST arguments
    if (
        hasattr(request, "arguments")
        and "apikey" in request.arguments
        and len(request.arguments["apikey"]) > 0
        and len(request.arguments["apikey"].strip()) > 0
    ):
        return request.arguments["apikey"]

    # Check Tornado and Flask headers
    if "X-Api-Key" in request.headers:
        return request.headers.get("X-Api-Key")

    # Check Tornado and Flask headers
    if "Authorization" in request.headers:
        header = request.headers.get("Authorization")
        if header.startswith("Bearer "):
            token = header.replace("Bearer ", "", 1)
            return token

    return None


def get_authorization_header(request):
    # Tornado and Flask headers
    return request.headers.get("Authorization")


def get_plugin_hash():
    from octoprint.plugin import plugin_manager

    plugin_signature = lambda impl: f"{impl._identifier}:{impl._plugin_version}"
    template_plugins = list(
        map(
            plugin_signature,
            plugin_manager().get_implementations(octoprint.plugin.TemplatePlugin),
        )
    )
    asset_plugins = list(
        map(
            plugin_signature,
            plugin_manager().get_implementations(octoprint.plugin.AssetPlugin),
        )
    )
    ui_plugins = sorted(set(template_plugins + asset_plugins))

    import hashlib

    plugin_hash = hashlib.sha1()
    plugin_hash.update(",".join(ui_plugins).encode("utf-8"))
    return plugin_hash.hexdigest()


def has_permissions(*permissions):
    """
    Determines if the current user (either from the session, api key or authorization
    header) has all of the requested permissions.

    Args:
        *permissions: list of all permissions required to pass the check

    Returns: True if the user has all permissions, False otherwise
    """
    logged_in = False

    try:
        if loginUserFromApiKey():
            logged_in = True
    except InvalidApiKeyException:
        pass  # ignored

    if not logged_in:
        loginUserFromAuthorizationHeader()

    flask.passive_login()
    return all(map(lambda p: p.can(), permissions))


def require_login_with(permissions=None, user_id=None):
    """
    Requires a login with the given permissions and/or user id.

    Args:
        permissions: list of all permissions required to pass the check
        user_id: required user to pass the check

    Returns: a flask redirect response to return if a login is required, or None
    """

    from octoprint.server import current_user, userManager

    login_kwargs = {"redirect": _flask.request.script_root + _flask.request.full_path}
    if (
        _flask.request.headers.get("X-Preemptive-Recording", "no") == "no"
        and userManager.has_been_customized()
    ):
        requires_login = False

        if permissions is not None and not has_permissions(*permissions):
            requires_login = True
            login_kwargs["permissions"] = ",".join([x.key for x in permissions])

        if user_id is not None and current_user.get_id() != user_id:
            requires_login = True
            login_kwargs["user_id"] = user_id

        if requires_login:
            return _flask.redirect(_flask.url_for("login", **login_kwargs))

    return None


def require_login(*permissions):
    """
    Returns a redirect response to the login view if the permission requirements are
    not met.

    Args:
        *permissions: a list of permissions required to pass the check

    Returns: a flask redirect response to return if a login is required, or None
    """
    from octoprint.server import userManager

    if not permissions:
        return None

    if (
        _flask.request.headers.get("X-Preemptive-Recording", "no") == "no"
        and userManager.has_been_customized()
    ):
        if not has_permissions(*permissions):
            return _flask.redirect(
                _flask.url_for(
                    "login",
                    redirect=_flask.request.script_root + _flask.request.full_path,
                    permissions=",".join([x.key for x in permissions]),
                )
            )

    return None


def validate_local_redirect(url, allowed_paths):
    """Validates the given local redirect URL against the given allowed paths.

    An `url` is valid for a local redirect if it has neither scheme nor netloc defined,
    and its path is one of the given allowed paths.

    Args:
        url (str): URL to validate
        allowed_paths (List[str]): List of allowed paths, only paths contained
            will be considered valid.

    Returns:
        bool: Whether the `url` passed validation or not.
    """
    from urllib.parse import urlparse

    parsed = urlparse(url)
    return parsed.scheme == "" and parsed.netloc == "" and parsed.path in allowed_paths



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2021 The OctoPrint Project - Released under terms of the AGPLv3 License"

import itertools
import os.path
import re


def _sfn_really_universal(name):
    from octoprint.util.text import sanitize

    ### taken from pathvalidate library

    _WINDOWS_RESERVED_FILE_NAMES = ("CON", "PRN", "AUX", "CLOCK$", "NUL") + tuple(
        f"{name:s}{num:d}"
        for name, num in itertools.product(("COM", "LPT"), range(1, 10))
    )
    _MACOS_RESERVED_FILE_NAMES = (":",)

    result = sanitize(name, safe_chars="-_.()[] ").replace(" ", "_")
    root, ext = os.path.splitext(result)
    if root.upper() in (_WINDOWS_RESERVED_FILE_NAMES + _MACOS_RESERVED_FILE_NAMES):
        root += "_"
    return root + ext


def sanitize_filename(name, really_universal=False):
    """
    Sanitizes the provided filename. Implementation differs between Python versions.

    Under normal operation, ``pathvalidate.sanitize_filename`` will be used, leaving the
    name as intact as possible while still being a legal file name under all operating
    systems.

    Behaviour can be changed by setting ``really_universal`` to ``True``. In this case,
    the name will be ASCII-fied, using ``octoprint.util.text.sanitize`` with
    safe chars ``-_.()[] `` and all spaces replaced by ``_``. This is the old behaviour.

    In all cases, a single leading ``.`` will be removed (as it denotes hidden files
    on *nix).

    Args:
        name:          The file name to sanitize. Only the name, no path elements.
        really_universal: If ``True``, the old method of sanitization will always
                          be used. Defaults to ``False``.

    Returns:
        the sanitized file name
    """
    from octoprint.util import to_unicode

    name = to_unicode(name)

    if name is None:
        return None

    if "/" in name or "\\" in name:
        raise ValueError("name must not contain / or \\")

    from pathvalidate import sanitize_filename as sfn

    if really_universal:
        result = _sfn_really_universal(name)
    else:
        result = sfn(name)

    return result.lstrip(".")


def get_dos_filename(
    input, existing_filenames=None, extension=None, whitelisted_extensions=None, **kwargs
):
    """
    Converts the provided input filename to a 8.3 DOS compatible filename. If ``existing_filenames`` is provided, the
    conversion result will be guaranteed not to collide with any of the filenames provided thus.

    Uses :func:`find_collision_free_name` internally.

    Arguments:
        input (string): The original filename incl. extension to convert to the 8.3 format.
        existing_filenames (list): A list of existing filenames with which the generated 8.3 name must not collide.
            Optional.
        extension (string): The .3 file extension to use for the generated filename. If not provided, the extension of
            the provided ``filename`` will simply be truncated to 3 characters.
        whitelisted_extensions (list): A list of extensions on ``input`` that will be left as-is instead of
            exchanging for ``extension``.
        kwargs (dict): Additional keyword arguments to provide to :func:`find_collision_free_name`.

    Returns:
        string: A 8.3 compatible translation of the original filename, not colliding with the optionally provided
            ``existing_filenames`` and with the provided ``extension`` or the original extension shortened to
            a maximum of 3 characters.

    Raises:
        ValueError: No 8.3 compatible name could be found that doesn't collide with the provided ``existing_filenames``.

    Examples:

        >>> get_dos_filename("test1234.gco")
        'test1234.gco'
        >>> get_dos_filename("test1234.gcode")
        'test1234.gco'
        >>> get_dos_filename("test12345.gco")
        'test12~1.gco'
        >>> get_dos_filename("Wölfe 🐺.gcode")
        'wolfe_~1.gco'
        >>> get_dos_filename("💚.gcode")
        'green_~1.gco'
        >>> get_dos_filename("test1234.fnord", extension="gco")
        'test1234.gco'
        >>> get_dos_filename("auto0.g", extension="gco")
        'auto0.gco'
        >>> get_dos_filename("auto0.g", extension="gco", whitelisted_extensions=["g"])
        'auto0.g'
        >>> get_dos_filename(None)
        >>> get_dos_filename("foo")
        'foo'
    """

    if input is None:
        return None

    input = sanitize_filename(input, really_universal=True)

    if existing_filenames is None:
        existing_filenames = []

    if extension is not None:
        extension = extension.lower()

    if whitelisted_extensions is None:
        whitelisted_extensions = []

    filename, ext = os.path.splitext(input)

    ext = ext.lower()
    ext = ext[1:] if ext.startswith(".") else ext
    if ext in whitelisted_extensions or extension is None:
        extension = ext

    return find_collision_free_name(filename, extension, existing_filenames, **kwargs)


def find_collision_free_name(filename, extension, existing_filenames, max_power=2):
    """
    Tries to find a collision free translation of "<filename>.<extension>" to the 8.3 DOS compatible format,
    preventing collisions with any of the ``existing_filenames``.

    First strips all of ``."/\\[]:;=,`` from the filename and extensions, converts them to lower case and truncates
    the ``extension`` to a maximum length of 3 characters.

    If the filename is already equal or less than 8 characters in length after that procedure and "<filename>.<extension>"
    are not contained in the ``existing_files``, that concatenation will be returned as the result.

    If not, the following algorithm will be applied to try to find a collision free name::

        set counter := power := 1
        while counter < 10^max_power:
            set truncated := substr(filename, 0, 6 - power + 1) + "~" + counter
            set result := "<truncated>.<extension>"
            if result is collision free:
                return result
            counter++
            if counter >= 10 ** power:
                power++
        raise ValueError

    This will basically -- for a given original filename of ``some_filename`` and an extension of ``gco`` -- iterate
    through names of the format ``some_f~1.gco``, ``some_f~2.gco``, ..., ``some_~10.gco``, ``some_~11.gco``, ...,
    ``<prefix>~<n>.gco`` for ``n`` less than 10 ^ ``max_power``, returning as soon as one is found that is not colliding.

    Arguments:
        filename (string): The filename without the extension to convert to 8.3.
        extension (string): The extension to convert to 8.3 -- will be truncated to 3 characters if it's longer than
            that.
        existing_filenames (list): A list of existing filenames to prevent name collisions with.
        max_power (int): Limits the possible attempts of generating a collision free name to 10 ^ ``max_power``
            variations. Defaults to 2, so the name generation will maximally reach ``<name>~99.<ext>`` before
            aborting and raising an exception.

    Returns:
        string: A 8.3 representation of the provided original filename, ensured to not collide with the provided
            ``existing_filenames``

    Raises:
        ValueError: No collision free name could be found.

    Examples:

        >>> find_collision_free_name("test1234", "gco", [])
        'test1234.gco'
        >>> find_collision_free_name("test1234", "gcode", [])
        'test1234.gco'
        >>> find_collision_free_name("test12345", "gco", [])
        'test12~1.gco'
        >>> find_collision_free_name("test 123", "gco", [])
        'test_123.gco'
        >>> find_collision_free_name("test1234", "g o", [])
        'test1234.g_o'
        >>> find_collision_free_name("test12345", "gco", ["/test12~1.gco"])
        'test12~2.gco'
        >>> many_files = ["/test12~{}.gco".format(x) for x in range(10)[1:]]
        >>> find_collision_free_name("test12345", "gco", many_files)
        'test1~10.gco'
        >>> many_more_files = many_files + ["/test1~{}.gco".format(x) for x in range(10, 99)]
        >>> find_collision_free_name("test12345", "gco", many_more_files)
        'test1~99.gco'
        >>> many_more_files_plus_one = many_more_files + ["/test1~99.gco"]
        >>> find_collision_free_name("test12345", "gco", many_more_files_plus_one)
        Traceback (most recent call last):
        ...
        ValueError: Can't create a collision free filename
        >>> find_collision_free_name("test12345", "gco", many_more_files_plus_one, max_power=3)
        'test~100.gco'

    """
    from octoprint.util import to_unicode

    filename = to_unicode(filename)
    extension = to_unicode(extension)

    if filename.startswith("/"):
        filename = filename[1:]
    existing_filenames = [
        to_unicode(x[1:] if x.startswith("/") else x) for x in existing_filenames
    ]

    def make_valid(text):
        return re.sub(
            r"\s+", "_", text.translate({ord(i): None for i in r".\"/\[]:;=,"})
        ).lower()

    filename = make_valid(filename)
    extension = make_valid(extension)
    extension = extension[:3] if len(extension) > 3 else extension

    full_name_format = "{filename}.{extension}" if extension else "{filename}"

    result = full_name_format.format(filename=filename, extension=extension)
    if len(filename) <= 8 and result not in existing_filenames:
        # early exit
        return result

    counter = 1
    power = 1
    prefix_format = "{segment}~{counter}"
    while counter < (10**max_power):
        prefix = prefix_format.format(
            segment=filename[: (6 - power + 1)], counter=str(counter)
        )
        result = full_name_format.format(filename=prefix, extension=extension)
        if result not in existing_filenames:
            return result
        counter += 1
        if counter >= 10**power:
            power += 1

    raise ValueError("Can't create a collision free filename")


def silent_remove(file):
    """
    Silently removes a file. Does not raise an error if the file doesn't exist.

    Arguments:
        file (string): The path of the file to be removed
    """

    try:
        os.remove(file)
    except OSError:
        pass
